__author__ = 'thibautlamadon'

import Formatter
import ast

class FortranFormat(Formatter.Formatter):

  def __init__(self):
    Formatter.Formatter.__init__(self)
    self.SUBSCRIPT_L = '('
    self.SUBSCRIPT_R = ')'
    self.POW = '**'
    self.COMMENT ="!"
    self.FILE_EXTENSION="f90"


  def declareTensor(self,t):
    return "double precision " + t.name + "(" + ",".join([ ":" for i in range(0,t.dim)]) + ")" +self.CR

  def declareResult(self,tp,sizes):
    return "double precision " + tp.name   + "("+ ",".join(sizes) + ")"

  def declareScalar(self,t):
    return "double precision " + t +self.CR

  def declareIndex(self,t):
    return "integer " + t +self.CR

  def visit_Num(self,node):
    self.content += str(node.n) +'D0'

  def declareCall(self,tp,sizes):
    args = [t.name for t in tp.tensors]
    args.sort()
    s = "function " + tp.name + "(" +",".join(args)
    if len(tp.scalars)>0:
      s += "," + ",".join(tp.scalars)
    s +=  "," + ",".join(sizes)

    s += ")" + self.CR + "implicit none"
    return s

  def getResName(self,tp):
    return tp.name

  def declareFunction(self,tp):
    self.RES_NAME = tp.name
    return Formatter.Formatter.declareFunction(self,tp)

  def declareSetResultTozero(self):
    return self.LHS + "= 0D0"

  def declareLoopIn(self,i,s):
    return "do " + i + " = 1," +  s
  def declareLoopOut(self,i):
    return "end do"
  def visit_Indicator(self,node):
    self.content += 'MERGE(1D0,0D0,'
    self.visit(node)
    self.content += ')'

  def declareModuleHeader(self,name,tensors):
    return "module " + name +self.CR + "contains" + self.CR

  def declareModuleFooter(self,name):
    return "end module "+ self.CR

  def func_footer(self,tp):
    return "end function" + self.CR




class Tensor:
  """ this stores a reference to a tensor, a name and its dimensions
  """

  def __init__(self,name,dim):
    self.dim   = dim
    self.name  = name
  def __eq__(self, other):
    return self.name==other.name
  def __hash__(self):
    ord3 = lambda x : '%.3d' % ord(x)
    return int(''.join(map(ord3, self.name)))
  def __str__(self):
    return self.name+ '[' + str(self.dim) + ']'
  def __lt__(self,other):
    return self.name < other.name

class Scalar:
  def __init__(self,name):
    self.name = name
  def __str__(self):
    self.name


class IndexLocation:
  """ this stores a reference to an index, a name and its dimensions
  """
  def __init__(self,tname,iname,dim):
    self.tname = tname
    self.iname = iname
    self.dim  = dim
  def __eq__(self, other):
    return (self.iname==other.iname) & (self.dim==self.dim) & (self.tname==self.tname)
  def __hash__(self):
    ord3 = lambda x : '%.3d' % ord(x)
    return int(''.join(map(ord3, self.iname + self.tname +  str(self.dim))))
  def __str__(self):
    return self.iname + ' : ' + str(self.dim) + ']'
  def __lt__(self,other):
    if (self.name == other.name):
      return self.dim < other.dim
    return self.name < other.name




__author__ = 'thibautlamadon'

import Formatter
import ast

class JuliaFormat(Formatter.Formatter):

  def __init__(self):
    Formatter.Formatter.__init__(self)
    self.SUBSCRIPT_L = '['
    self.SUBSCRIPT_R = ']'
    self.POW = '.^'
    self.COMMENT ="#"
    self.FILE_EXTENSION="jl"
    self.BinOpMid = { "Mult":".*", "Add":".+", "Div":"./", "Sub":".-" }
    self.Compare = { "Gt":".>","GtE":".>=","Lt":".<","LtE":".<=" }
    self.INDEX_LINEARISE =1

  def func_header(self,tp):
    return ""

  def func_footer(self,tp):
    return "end"

  def declareTensor(self,t):
    return ""

  def declareResult(self,tp,sizes):
    return self.RES_NAME + " = zeros("+ ",".join(sizes) + ")"

  def declareIndex(self,t):
    return ""

  def declareCall(self,tp,sizes):
    args = [t.name for t in tp.tensors]
    args.sort()
    s = "function " + tp.name + "(" +",".join(args)
    if len(tp.scalars)>0:
      s += "," + ",".join(tp.scalars)
    s += ")"

    s += self.CR

    # extract sizes from inputs
    for ss in sizes :
      tname,dim = tp.getIndexSize(ss)
      s += ss + " = size(" + tname + ")[" + str(dim+1) + "]" + self.CR

    s += "@inbounds begin"

    return s

  def getResName(self,tp):
    return tp.name

  def declareFunction(self,tp):
    self.RES_NAME = "Res" #tp.name
    return Formatter.Formatter.declareFunction(self,tp)

  def declareLoopIn(self,i,size):
    return "for " + i + " in 1:" +  size

  def declareLoopOut(self,i):
    return "end"

  def visit_Indicator(self,node):
    self.content += '('
    self.visit(node)
    self.content += ')'

  def func_footer(self,tp):
    s = "end" + self.CR
    s += "return " + self.RES_NAME + self.CR+ "end" + self.CR
    return s

  def declareModuleHeader(self,name,tensors):
    rr = "module " + name +self.CR
    rr += "export " + ",".join(tensors)
    return rr


  def declareModuleFooter(self,name):
    return self.CR + "end"+ self.CR


"""
This is relatively basic stuff
you write an expression of the sort

A[x1,x2] * B[x2]

I collect the indexes, here x1,x2
I generate the code to wrap the operation as a sum.

can generate fortran, c++, or anything else that can be useful

"""


import ast
import tensorparser.FormatterJulia
format = tensorparser.FormatterJulia.JuliaFormat()
tp = tensorparser.TensorParser("T_hire      " , "I(S[x,y2] - Z[z1] >= 0 ) * G[z1] * V[y2]","x")
print format.declareFunction(tp)


#process("T_P0_1     " , "max(0,S[x1,y] - S[x1,y1]-Z[z1]) * G[z1] * H[x1,y1]","y")
#process("T_P0_0     " , "max(0,S[x1,y]-Z[z1]) * G[z1] * U[x1]","y")
#process("T_W0       " , "max(0,S[x,y1]-Z[z1]) * G[z1] * V[y1] ","x")
#process("T_H_j2j    " , "I(S[x,y] -Z[z1]  > S[x,y2]       ) * G[z1] * H[x,y2] * V[y]","x,y")
#process("T_H_j2j_out" , "I(S[x,y]        < S[x,y2] -Z[z1] ) * G[z1] * V[y2]","x,y")
#process("T_H_u2e    " , "I(S[x,y] -Z[z1] >=0) * G[z1] * U[x] * V[y]","x,y")
#process("T_V        " , "H[x,y]","y")
#process("T_U        " , "H[x,y]","x")
#process("T_pr0_0    " , "I(S[x,y] -Z[z1]  >0 ) * G[z1] * U[x] ","y")
#process("T_pr0_1    " , "I(S[x,y] -Z[z1]  > S[x,y2]       ) * G[z1] * H[x,y2] ","y")
#process("T_pr1      " , "I(S[x,y]        < S[x,y2] -Z[z1] ) * G[z1] * V[y2]","y")

#process("T_hire      " , "I(S[x,y2] - Z[z1] >= 0 ) * G[z1] * V[y2]","x")
#
#
#
#
#process("T_sep      " , "I(S[x,y2] < 0 ) * H[x,y2]","x")
#
#process("T_fill0      " , "I(S[x2,y] - Z[z1] >= 0 ) * G[z1] * U[x2]","y")
#process("T_fill1      " , "I(S[x2,y] - Z[z1] >= S[x2,y2] ) * G[z1] * H[x2,y2]","y")
#process("T_leave      " , "I(S[x,y2] - Z[z1] > S[x,y] )   * V[y2] * G[z1] * H[x,y]","y")



__author__ = 'thibautlamadon'

"""
 Mother class to generate output of the tensor.
 Language specific implementation should extend that class.

"""

import ast
import re

class Formatter(ast.NodeVisitor):

  def __init__(self):
    self.CR = "\n"
    self.content = ""
    self.BinOpMid = { "Mult":"*", "Add":"+", "Div":"/", "Sub":"-" }
    self.CALL2 = { "max":"max"}
    self.Functions = { "log":"log" }
    self.Compare = { "Gt":">","GtE":">=","Lt":"<","LtE":"<=" }
    self.SUBSCRIPT_L = '['
    self.SUBSCRIPT_R = ']'
    self.POW = '^'
    self.RES_NAME = 'R'
    self.LHS = 'R'
    self.indexAsSizes = False
    self.COMMENT = "//"
    self.FILE_EXTENSION = "txt"
    self.INDEX_LINEARISE = 0 # 0 for multidimensional, 1 for linear starting at 0, 2 for starting at 1
    self.tp = None
    pass

  def visit_wrap(self,n):
    self.generic_visit(ast.Tuple(elts=[n]))

  def visit_Module(self, node):
    """ Called on the beginning of the visit
    """
    self.content = ""
    self.generic_visit(node)

  def generic_visit(self, node):
    ast.NodeVisitor.generic_visit(self, node)

  def visit_Name(self, node):
    self.content += node.id

  def visit_Subscript(self, node):
    AA = node.value.id
    if (AA=='R'): AA = self.RES_NAME
    self.content +=  AA + self.SUBSCRIPT_L
    ast.NodeVisitor.generic_visit(self, node.slice)
    self.content += self.SUBSCRIPT_R

  def visit_Tuple(self,node):
    """ generates the multi-dimensional array, it would a good idea
        to allow to generate a linear index x1 + nx*( y2 - 1)
    """
    IS = [e.id for e in node.elts]
    if (self.INDEX_LINEARISE==0):
      self.content += ",".join(IS)
    else:
      prev= ""
      for e in reversed(IS):
        if (prev==""):
          EE = e
        else:
          EE = e + " + " + self.tp.getIndexName(e) + " * (" + EE + "-1)"
        prev = e
      self.content += EE


  def visit_Call(self,node):
    if (node.func.id=='I'):
      for l in node.args: self.visit_Indicator(l)
    elif (node.func.id in self.CALL2):
      self.content += node.func.id + '('
      self.visit_wrap(node.args[0])
      self.content +=  ','
      self.visit_wrap(node.args[1])
      self.content +=  ')'
    else:
      self.content += node.func.id + '('
      for l in node.args: self.visit_wrap(l)
      self.content +=  ')'


  def visit_BinOp(self,node):
    nn = node.op.__class__.__name__

    if (nn=='Pow'):
      self.visit_Pow(node.left,node.right)
    elif (nn in self.BinOpMid.keys()):
      self.visit_wrap(node.left)
      self.content += self.BinOpMid[nn]
      self.visit_wrap(node.right)

  def visit_Compare(self,node):
    self.visit_wrap(node.left)

    nn = node.ops[0].__class__.__name__
    if (nn in self.Compare.keys()):
      self.content += self.Compare[nn]
    else:
      self.content += nn

    self.visit_wrap(node.comparators[0])

  def visit_Indicator(self,node):
    self.content += '1['
    self.visit(node)
    self.content += ']'

  def formatTensor(self,tp):
    self.content = ""
    self.generic_visit(tp.Et)
    print self.content

  def visit_Num(self,node):
    self.content += str(node.n)

  def visit_Pow(self,x1,x2):
    self.content += '('
    self.visit_wrap(x1)
    self.content += ')'+self.POW + '('
    self.visit_wrap(x2)
    self.content += ')'

  def declareExtractIndexRange(self,iname,tname,dim):
    return "int " + iname + " = dim(" + tname + ")[" + str(dim) + "]"

  def declareFunction(self,tp):
    """
    generate the full function
    """

    self.tp = tp

    # extract sizes
    sizes = tp.getSizesSet()
    sp = lambda x: "".join([ " " for i in range(0,x)]) # function that indents

    # get result
    self.LHS = self.RES_NAME + self.SUBSCRIPT_L + ",".join(tp.index_out) + self.SUBSCRIPT_R

    s = ""
    s += self.func_header(tp) + self.CR

    # add a commenter with the formula
    s += self.addComment( "Generated tensor:") + self.CR
    s += self.addComment( tp.name + " " + tp.Eraw + " | " + tp.Vraw ) + self.CR
    # print function signature
    s += self.declareCall(tp,sizes) + self.CR

    for t in tp.index_out:
      s += self.declareIndex(t)
    for t in tp.index_rest:
      s += self.declareIndex(t)
    for t in tp.tensors :
      s += self.declareTensor(t)
    for t in tp.scalars:
      s += self.declareScalar(t)

    # declare result
    s += self.declareResult(tp,[tp.getIndexName(i) for i in tp.index_out]) + self.CR

    # create loops or out indexes
    indent = 0
    for i in tp.index_out:
      s += sp(indent) + self.declareLoopIn(i,tp.getIndexName(i)) + self.CR
      indent+=1

    # initialing the result to 0
    s += sp(indent) + self.declareSetResultTozero() + self.CR

    for i in tp.index_rest:
      s += sp(indent) + self.declareLoopIn(i,tp.getIndexName(i)) + self.CR
      indent+=1

    # append formula
    s += sp(indent) + self.declareFormula(tp) + self.CR

    for i in tp.index_out:
      indent-=1
      s += sp(indent) + self.declareLoopOut(i)+ self.CR

    for i in tp.index_rest:
      indent-=1
      s += sp(indent) + self.declareLoopOut(i) + self.CR

    s += self.func_footer(tp)

    return s

  def declareTensor(self,t):
    return ""

  def declareResult(self,n):
    return ""

  def declareScalar(self,t):
    return ""

  def declareFormula(self,tp):
    self.visit(tp.Et)
    et_str = self.content
    return self.LHS + " =  " + self.LHS + self.BinOpMid["Add"] + et_str

  def declareLoopIn(self,i,size):
    return "for " + i + " = 1:" +  self.getIndexName(i)
  def declareLoopOut(self,i):
    return "end"
  def declareSetResultTozero(self):
    return self.LHS + "= 0"
  def addComment(self,s):
    return self.COMMENT + ' ' + s
  def getFileExtension(self):
    return "." + self.FILE_EXTENSION
  def declareModuleHeader(self,name,tensors=set()):
    return "Module file header"
  def declareModuleFooter(self,name):
    return "Module file footer"

  def func_header(self,tp):
      return ""

  def func_footer(self,tp):
    return "" + self.CR



from setuptools import setup

setup(name='tensorparser',
      version='0.1',
      description='Generate code from tensor expressions',
      url='http://github.com/storborg/funniest',
      author='Thibaut Lamadon',
      author_email='thibaut.lamadon@gmail.com',
      license='MIT',
      packages=['tensorparser'],
      scripts=['bin/tensorparser'],
      zip_safe=False)


import ast
from elements import *
import re

"""
I need to extract:
 - the Arrays and their dimensions
 - the scalars
 - the list of output indexes
 - the list of all indexes
 - the common indexes so that dimension can be passed to the routine

 help: http://greentreesnakes.readthedocs.org/en/latest/nodes.html#control-flow
"""

class IndexCollector(ast.NodeVisitor):
  ''' parses expressions like a[x,y,z] and extracts
      the list of indexes x,y,z. We also get the
      order
  '''
  def __init__(self):
    ast.NodeVisitor.__init__(self)
    self.indexes = []

  def generic_visit(self, node):
    ast.NodeVisitor.generic_visit(self, node)

  def visit_Name(self, node):
    self.indexes.append(node.id)

class IndexFinder(ast.NodeVisitor):
  '''
  Extracts the different indexes present in the tensor expression,
  as well as where they appear
  '''

  def __init__(self):
    ast.NodeVisitor.__init__(self)
    self.tensors = set()
    self.indexes = set()
    self.index_loc = dict()
    self.scalars = set()

  def visit_Module(self, node):
    """ Called on the beginning of the visit
    """
    self.tensors = set()
    self.indexes = set()
    self.scalars = set()
    self.index_loc = set()
    self.generic_visit(node)

  def generic_visit(self, node):
    ast.NodeVisitor.generic_visit(self, node)

  def visit_Subscript(self,node):
    """
    when we find an Array, we need
      - name
      - dimension
      - list of indexes and their locations
    """
    cc = IndexCollector()
    ast.NodeVisitor.generic_visit(cc, node.slice)

    # store the tensor
    self.tensors.add(Tensor(node.value.id,len(cc.indexes)))
    # store the indexes
    self.indexes = set(cc.indexes) | set(self.indexes)

    # store each location
    i = 0
    for ii in cc.indexes:
      self.index_loc.add(IndexLocation(node.value.id,ii,i))
      i+=1


  def visit_Name(self, node):
    self.scalars.add(node.id)

  def visit_Call(self,node):
    for l in node.args: ast.NodeVisitor.generic_visit(self, ast.Tuple(elts=[l]))

class TensorParser:
  """
   very important class that extracts information from the tensor expression.
   From this class, the formatter should be able to generate the code for the module.
  """

  def __init__(self,name,E,vars):
    self.Eraw = E
    self.Vraw = vars
    self.expr = E
    self.Et = ast.parse(E)
    self.vars = vars.split(',')
    self.name = name.strip()

    self.index_out   = vars.split(',')

    # extract the other indexes
    index_finder   = IndexFinder()
    index_finder.visit(self.Et)
    self.index_rest = index_finder.indexes - set(self.index_out)

    # extract tensors and scalars
    self.tensors = index_finder.tensors
    self.scalars = index_finder.scalars
    self.index_loc = index_finder.index_loc

    # extract unique indexes
    # we need to remove the number at the end of each index

  def __str__(self):
    s = 'I(' + ','.join(self.index_rest) + '|' + ','.join(self.index_out) + ')'
    s += ',T(' + ','.join([ str(t) for t in self.tensors]) + ')'
    s += ',S(' + ','.join([ str(t) for t in self.scalars]) + ')'
    return s
#    print "index others: ",self.index_rest
#    print "index tensors:",self.tensors
#    print "index scalars:",self.scalars

  def getIndexSize(self,iname):
    for ii in self.index_loc:
      if ( iname == self.getIndexName(ii.iname) ):
        return ii.tname,ii.dim
    raise Exception(iname + " not in location list!")


  def getIndexName(self,index):
    """
    if index is x1, returns nx,
    in other words, remove the numbers at the end

    """
    return "n" + re.search("(.*[a-zA-Z])[0-9]*$", index).group(1)

  def getSizesSet(self , withn=False):
    if (withn):
      return set([self.getIndexName(ii) for ii in (self.index_rest| set(self.index_out) )])
    else:
      return set([self.getIndexName(ii) for ii in (self.index_rest| set(self.index_out) )])


from unittest import TestCase
from tensorparser import TensorParser
from tensorparser.elements import IndexLocation
import json

__author__ = 'thibautlamadon'

class TestTensorParser(TestCase):

  def  test_extract_index(self):
    tp = TensorParser("T_hire      " , "I(S[x,y2] - Z[z1] >= 0 ) * G[z1] * V[y2]","x")
    self.assertEqual(tp.index_out, ['x'])
    self.assertTrue("y2" in tp.index_rest)
    self.assertTrue("z1" in tp.index_rest)
    self.assertTrue("nz" in tp.getSizesSet())
    self.assertTrue("nx" in tp.getSizesSet())
    self.assertTrue("ny" in tp.getSizesSet())

  def  test_extract_index(self):
    """

    """
    tp = TensorParser("T_hire      " , "I(S[x,y2] - Z[z1] >= 0 ) * G[z1] * V[y2]","x")
    tname,dim = tp.getIndexSize('nz')
    testbool  = ((tname=="S") & (dim==1)) | ((tname=="G") & (dim==0))
    self.assertTrue(testbool)

  def test_index(self):
    i1 = IndexLocation('T','x',1)
    self.assertTrue(i1.iname=='x')


  def test_json(self):
    with open("../examples/atensor.json") as data_file:
      data = json.load(data_file)





